# define basic image info dict keys
OS_FAMILY = "os_family"
OS_NAME = "os_name"
OS_ARCHI = "os_archi"
REPO_TAGS = "repo_tags"
IMG_ID = "image_id"
IMG_NAME = "image_name"
IMG_CREATION_TIME = "creation_time"

# define results parsed dict keys
TARGET_NAME = "target_name"
TARGET_CLASS = "target_class"
TARGET_TYPE = "target_type"
TARGET_VULNERABILITY_DF = "target_vul_df"

# define vulnerability dataframe column name
LIBRARY = "library"
VULNERABILITY_ID = "vulnerability_id"
VULNERABILITY_URL = "vulnerability_url"
SEVERITY = "severity"
STATUS = "status"
INSTALLED_VER = "installed_version"
FIXED_VER = "fixed_version"
DESCRIPTION = "description"
CVSS_V3_SCORE = "cvss_v3_score"
CVSS_V3_VECTOR = "cvss_v3_vector"



"""
image_vul_parser
================

This module provides a class to parse the raw JSON report which is generated by Trivy.

Author:
    Pengfei liu

Date:
    2025-02-25
"""

# Standard library imports
import json
from typing import List

# Third party imports
import pandas as pd
from pandas import DataFrame

# sherlock imports
from iv_sherlock.conf.constant import (
    CVSS_V3_SCORE,
    CVSS_V3_VECTOR,
    DESCRIPTION,
    FIXED_VER,
    IMG_CREATION_TIME,
    IMG_ID,
    IMG_NAME,
    INSTALLED_VER,
    LIBRARY,
    OS_ARCHI,
    OS_FAMILY,
    OS_NAME,
    REPO_TAGS,
    SEVERITY,
    STATUS,
    TARGET_CLASS,
    TARGET_NAME,
    TARGET_TYPE,
    TARGET_VULNERABILITY_DF,
    VULNERABILITY_ID,
    VULNERABILITY_URL,
)


def print_keys_as_tree(in_dict: dict, indent=0):
    """
    Recursively prints the keys of a nested dictionary as a tree structure
    :param in_dict: The input dictionary
    :param indent: the number of spaces to indent each level
    :return:
    """
    for key, value in in_dict.items():
        print("  " * indent + str(key))  # Print the key with indentation
        if isinstance(value, dict):
            # If the value is a dictionary, recurse into it
            print_keys_as_tree(value, indent + 1)


class ImageVulParser:
    """
    This class represents an image_vulnerability_parser that parses the raw image vulnerability
    report that is generated by trivy in JSON format.

    :param raw_report_path:  The file path which contains the raw image vulnerability report.
    :type raw_report_path: Str

    :ivar raw_data: The dict which corresponds to the json file.
    :vartype raw_data: dict
    """

    def __init__(self, raw_report_path: str):
        """
        Initialize the image vulnerability parser
        :param raw_report_path: The JSON file path of the raw image vulnerability report.
        """
        with open(raw_report_path, "r", encoding="utf-8") as file:
            self.raw_data:dict = json.load(file)

    def get_raw_data(self) -> dict:
        """
        This method returns the raw dictionary of the trivy report.
        :return:
        """
        return self.raw_data

    def get_available_attributes(self) -> list:
        """
        This method returns the available attributes of the raw trivy report.
        :return: A list of available attributes.
        """
        return list(self.raw_data.keys())

    def build_image_vul_report(self) -> tuple[dict, list]:
        """
        This function builds an image vulnerability analysis summary report based on the raw data. It
        contains the basic image information and the list of image CVE
        :return:
        """
        img_basic_info = self.get_image_basic_info()
        cve_list = self.parser_cve_results()
        return img_basic_info, cve_list

    def parser_cve_results(self) -> List[dict]:
        """
        This function parses the Results section of the raw report. The results section contains a list of
        `Target`. A Target has the below attributes:
        - target_name: for example, redis (debian 12.7)
        - target_class: os-pkgs
        - target_type: debian
        - vulnerabilities: a list of dict which contains the details of each vulnerability
        For each target, we generate a parsed_target which contains:
        - target_name: for example, redis (debian 12.7)
        - target_class: os-pkgs
        - target_type: debian
        - target_vul_df: a pandas data frame to represent vulnerabilities(each row is a CVE)
        :return:
        """
        parsed_targets = []
        targets = self.raw_data.get("Results")
        # if the target is not empty, parse each target
        if targets:
            for target in targets:
                parsed_target = {}
                target_name = target.get("Target")
                target_class = target.get("Class")
                target_type = target.get("Type")
                vulnerabilities = target.get("Vulnerabilities")
                # build a pandas data frame to represent vulnerabilities(each row is a CVE)
                vul_pdf = self.parser_target_vulnerabilities(vulnerabilities)
                parsed_target[TARGET_NAME] = target_name
                parsed_target[TARGET_CLASS] = target_class
                parsed_target[TARGET_TYPE] = target_type
                parsed_target[TARGET_VULNERABILITY_DF] = vul_pdf
                parsed_targets.append(parsed_target)
        return parsed_targets

    @staticmethod
    def parser_target_vulnerabilities(vulnerabilities: List[dict]) -> DataFrame | None:
        """
        This function takes a list of vulnerabilities (raw.Results.Targets[i].Vulnerabilities), then parses them
        into a dataframe. It contains below columns:
        - library: name of the package which contains vulnerabilities
        - vulnerability_id: The id of the vulnerability(cve)
        - severity: The severity of the vulnerability(cve)
        - status: The status of the vulnerability(cve)
        - installed_version: The installed version of the package.
        - fixed_version: If a fix exists, in which version the fix will be applied
        - description: Description of the vulnerability(cve)
        - cvss_v3_score: The cvss score of the vulnerability(cve). Various organizations provide this score,
                 in this implementation, we took the first available, for some cve, this column maybe null.
        - cvss_v3_vector: The cvss vector of the vulnerability(cve)
        :param vulnerabilities:
        :return:
        """
        vul_list: list = []
        if vulnerabilities:
            for vul in vulnerabilities:
                vul_attrs = {}
                library = vul.get("PkgName")
                vul_id = vul.get("VulnerabilityID")
                vul_url = vul.get("PrimaryURL")
                severity = vul.get("Severity")
                status = vul.get("Status")
                installed_version = vul.get("InstalledVersion")
                fixed_version = vul.get("FixedVersion")
                description = vul.get("Title")
                # cvss_v3_score = vul.get("CVSS").get("nvd").get("V3Score")
                # cvss_v3_vector = vul.get("CVSS").get("nvd").get("V3Vector")
                cvss_v3_score, cvss_v3_vector = ImageVulParser.parse_cvss(
                    vul.get("CVSS")
                )
                vul_attrs[LIBRARY] = library
                vul_attrs[VULNERABILITY_ID] = vul_id
                vul_attrs[VULNERABILITY_URL] = vul_url
                vul_attrs[SEVERITY] = severity
                vul_attrs[STATUS] = status
                vul_attrs[INSTALLED_VER] = installed_version
                vul_attrs[FIXED_VER] = fixed_version
                vul_attrs[CVSS_V3_SCORE] = cvss_v3_score
                vul_attrs[CVSS_V3_VECTOR] = cvss_v3_vector
                vul_attrs[DESCRIPTION] = description
                vul_list.append(vul_attrs)
            pdf = pd.DataFrame(vul_list)
            return pdf
        else:
            return None

    @staticmethod
    def parse_cvss(cvss:dict)->(float, str):
        """
        This function takes the cvss section (raw.Results.Targets[i].Vulnerabilities[i].CVSS), and get the first
        CVSS available record, parse them and return score and vector. If no cvss record exists, return none, none.
        :param cvss: A dict which contains CVSS record issued by different organizations
        :return: a tuple of cvss score and cvss vector, if nothing found return None, None
        """
        if cvss:
            cvss_org = next(
                iter(cvss), None
            )  # Get the first organization name, if available
            cvss_v3_score = cvss.get(cvss_org, {}).get("V3Score")
            cvss_v3_vector = cvss.get(cvss_org, {}).get("V3Vector")
        else:
            cvss_v3_score = None
            cvss_v3_vector = None
        return cvss_v3_score, cvss_v3_vector

    def get_image_basic_info(self) -> dict:
        """
        This function return a dictionary of the basic info of the image, such as the below example:
        - image_name: redis
        - repo_tag: redis:latest
        - creation_time: 2024-09-13T09:13:14.153929471+02:00
        - os_family: debian
        - os_name: 12.7
        - os_architecture: amd64
        - image_id: sha256:590b81f2fea1af9798db4580a6299dafba020c2f5dc7d8d734663e7fa5299ca0
        :return:
        """
        # get some basic info from the metadata
        basic_info = self.parse_image_metadata()

        # add image name to dict
        basic_info[IMG_NAME] = self.raw_data.get("ArtifactName")

        # add creation time to dict
        basic_info[IMG_CREATION_TIME] = self.raw_data.get("CreatedAt")
        return basic_info

    def parse_image_metadata(self) -> dict:
        """
        This function parse the metadata of the image and return a dictionary of the basic infomation of the image,
        such as the below example:
        - os_family: debian
        - os_name: 12.7
        - os_architecture: amd64
        - repo_tag: redis:latest
        - image_id: sha256:590b81f2fea1af9798db4580a6299dafba020c2f5dc7d8d734663e7fa5299ca0

        :return:
        """
        result: dict = {}
        metadata = self.raw_data.get("Metadata")
        os_family: str = metadata.get("OS").get("Family")
        os_name: str = metadata.get("OS").get("Name")
        os_archi: str = metadata.get("ImageConfig").get("architecture")
        repo_tags: List = metadata.get("RepoTags")
        image_id: str = metadata.get("ImageID")
        result[OS_FAMILY] = os_family
        result[OS_NAME] = os_name
        result[OS_ARCHI] = os_archi
        result[REPO_TAGS] = repo_tags
        result[IMG_ID] = image_id
        return result



import json
import os
import pytest
import tempfile
import toml
from iv_sherlock.config_manager import Config


@pytest.fixture
def custom_config():
    return {'report': {'source_path': '/tmp/iv_sherlock/data', 'out_path': '/tmp/iv_sherlock/data/tmp',
                       'default_encoding': 'ASCII', 'export_pdf': True}, 'cvss': {'score_mapping': ''}}


@pytest.fixture
def score_mapping_dict():
    return {'AV': {'N': 0.85, 'A': 0.62}, 'AC': {'L': 0.77, 'H': 0.44}, 'PR': {'N': 0.85, 'L': 0.62, 'H': 0.27},
            'UI': {'N': 0.85, 'R': 0.62}, 'S': {'U': 1.0, 'C': 1.08}, 'C': {'H': 0.56, 'L': 0.22, 'N': 0.0},
            'I': {'H': 0.56, 'L': 0.22, 'N': 0.0}, 'A': {'H': 0.56, 'L': 0.22, 'N': 0.0}}


@pytest.fixture
def score_mapping_str(score_mapping_dict):
    return json.dumps(score_mapping_dict)


@pytest.fixture
def default_config(score_mapping_str):
    return {
        'report': {'source_path': '/home/pliu/git/ImageVulnAnalyzer/src/iv_sherlock/data',
                   'out_path': '/home/pliu/git/ImageVulnAnalyzer/src/iv_sherlock/data/tmp',
                   'default_encoding': 'utf-8', 'export_pdf': False},
        'cvss': {
            'score_mapping': score_mapping_str }}


@pytest.fixture
def temp_config_files(custom_config):
    """Fixture to create temporary config files for testing."""
    temp_dir = tempfile.TemporaryDirectory()
    default_config_path = os.path.join(temp_dir.name, "default_config.toml")
    custom_config_path = os.path.join(temp_dir.name, "custom_config.toml")

    # Create a user config file with overridden values
    with open(custom_config_path, "w") as f:
        toml.dump(custom_config, f)

    yield temp_dir.name, default_config_path, custom_config_path

    temp_dir.cleanup()  # Cleanup after test


@pytest.fixture
def config(temp_config_files):
    """Fixture to initialize the Config instance with test files."""
    temp_dir, default_config_path, custom_config_path = temp_config_files
    config_instance = Config(custom_config_path)
    return config_instance

@pytest.fixture
def config_without_custom_input():
    config_instance = Config()
    return config_instance


def test_default_config_loading(config, default_config):
    """Test if default config values are loaded correctly."""
    expected_default_config = default_config
    print(config._default_config)
    assert config._default_config == expected_default_config


def test_custom_config_loading(config, custom_config):
    """Test if custom config values are loaded correctly."""
    expected_custom_config = custom_config
    print(config._custom_config)
    assert config._custom_config == expected_custom_config


def test_get_input_report_path(config):
    """Test if the input report path is correct."""
    expected_in_path = "/tmp/iv_sherlock/data"
    actual_in_path = config.get_input_report_path()
    print(actual_in_path)
    assert actual_in_path == expected_in_path


def test_get_output_report_path(config):
    """Test if the output report path is correct."""
    expected_out_path = "/tmp/iv_sherlock/data/tmp"
    actual_out_path = config.get_output_report_path()
    print(actual_out_path)
    assert actual_out_path == expected_out_path


def test_get_export_pdf(config):
    """Test if the value export pdf is correct."""
    expected_pdf_val = True
    actual_pdf = config.get_export_pdf()
    print(actual_pdf)
    assert actual_pdf == expected_pdf_val


def test_get_default_encoding(config):
    """Test if the value default encoding is correct."""
    expected_encoding = "ASCII"
    actual_encoding = config.get_default_encoding()
    print(actual_encoding)
    assert actual_encoding == expected_encoding


def test_get_cvss_vector_value_mapping(config,score_mapping_dict):
    """Test if the cvss vector value mapping is correct."""
    actual_cvss = config.get_cvss_vector_value_mapping()
    print(actual_cvss)
    assert actual_cvss == score_mapping_dict


def test_get_input_report_path_without_custom_input(config_without_custom_input):
    """Test if the input report path is correct. As no custom config, the default config is used"""
    expected_in_path = "/home/pliu/git/ImageVulnAnalyzer/src/iv_sherlock/data"
    actual_in_path = config_without_custom_input.get_input_report_path()
    print(actual_in_path)
    assert actual_in_path == expected_in_path


def test_get_output_report_path_without_custom_input(config_without_custom_input):
    """Test if the output report path is correct."""
    expected_out_path = "/home/pliu/git/ImageVulnAnalyzer/src/iv_sherlock/data/tmp"
    actual_out_path = config_without_custom_input.get_output_report_path()
    print(actual_out_path)
    assert actual_out_path == expected_out_path


def test_get_export_pdf_without_custom_input(config_without_custom_input):
    """Test if the value export pdf is correct."""
    expected_pdf_val = False
    actual_pdf = config_without_custom_input.get_export_pdf()
    assert actual_pdf == expected_pdf_val


def test_get_default_encoding_without_custom_input(config_without_custom_input):
    """Test if the value default encoding is correct."""
    expected_encoding = "utf-8"
    actual_encoding = config_without_custom_input.get_default_encoding()
    print(actual_encoding)
    assert actual_encoding == expected_encoding


def test_get_cvss_vector_value_without_custom_input(config_without_custom_input,score_mapping_dict):
    """Test if the cvss vector value mapping is correct."""
    actual_cvss = config_without_custom_input.get_cvss_vector_value_mapping()
    print(actual_cvss)
    assert actual_cvss == score_mapping_dict

def test_singleton_behavior(temp_config_files):
    """Test that multiple instances reference the same config."""
    _,_,custom_config_path = temp_config_files
    # create config1 without a custom config
    config1 = Config()
    config1_expected_encoding = "utf-8"
    config1_actual_encoding = config1.get_default_encoding()
    assert config1_expected_encoding == config1_actual_encoding
    # create config2 with a custom config
    config2 = Config(custom_config_path)
    config2_expected_encoding = "ASCII"
    config2_actual_encoding = config2.get_default_encoding()
    assert config2_expected_encoding == config2_actual_encoding
    # as config1 and config2 point to the same object, so the value of config1 is updated too.
    config1_encoding_after_config2 = config1.get_default_encoding()
    assert config1_encoding_after_config2 == config2_expected_encoding
    assert config2_actual_encoding == config1_encoding_after_config2 # Singleton should share state




"""
report_generator
===============

This module provides a class to generate html and pdf reports based on the results of image_vul_parser.

Author:
    Pengfei liu

Date:
    2025-02-25
"""
import sys
# Standard library imports
from pathlib import Path
from typing import List

# Third party imports
import pandas as pd
from jinja2 import Environment, FileSystemLoader, TemplateNotFound

# sherlock imports

from iv_sherlock.conf.constant import (
    IMG_NAME,
    TARGET_VULNERABILITY_DF,
)
from iv_sherlock.config_manager import Config


class ReportGenerator:
    """
    Represents a report generator. It can generate a vulnerability report for each image, and a general
    report which summery the vulnerability of all existing images.

    :param template_dir_path: The directory path which contains the report HTML templates.
    :type template_dir_path: str
    :param output_path: The directory path which hosts the generated reports.
    :type output_path: str


    :ivar output_path: The directory path which hosts the generated reports.
    :vartype output_path: str

    """
    def __init__(self, config: Config):
        template_dir_path = Config.get_report_template_dir_path()
        if self._is_path_valid(template_dir_path):
            environment = Environment(loader=FileSystemLoader(template_dir_path))
            self.template = environment.get_template("vul_report.html")
            self.general_template = environment.get_template("gen_report.html")
            self.vul_simple_template = environment.get_template("vul_simple.html")
            self.end_general_template = environment.get_template("end_template.html")
        else:
            raise ValueError("The provided template dir path is not valid.")
        self.output_path = config.get_output_report_path()
        self.encoding = config.get_default_encoding()


    @staticmethod
    def _is_path_valid(dir_path: str):
        return Path(dir_path).is_dir()

    def generate_report(self, image_info: dict, cve_list: List[dict]):
        for cve in cve_list:
            vul_df: pd.DataFrame = pd.DataFrame(cve[TARGET_VULNERABILITY_DF])
            vul_html = vul_df.to_html(index=False)
            cve["target_vul_html"] = vul_html
        content = self.template.render(image_info=image_info, cve_list=cve_list)
        raw_image_name = image_info[IMG_NAME]
        image_name = self._normalize_image_name(raw_image_name)
        report_path = Path(self.output_path) / f"{image_name}_report.html"
        try:
            with open(report_path.as_posix(), "w+", encoding=self.encoding) as f:
                f.write(content)
        except FileNotFoundError:
            print(f"Can't create report file {report_path}, make sure the {self.output_path} directory exists.")
            sys.exit(1)

    def initiate_general_report(self, image_list):
        image_infos = []
        targets = 0
        cve_count_by_status_list = []
        for base_info, cve_list in image_list:
            targets = 0
            for cve in cve_list:
                targets = targets + 1
                vul_df: pd.DataFrame = pd.DataFrame(cve[TARGET_VULNERABILITY_DF])
                if len(vul_df) != 0:
                    cve_count_by_status = (
                        vul_df.groupby("severity").size().reset_index(name="CVE_Count")
                    )
                    cve_count_by_status_list.append(cve_count_by_status)

            combined_df = pd.concat(cve_count_by_status_list)
            cve_count_by_status = combined_df.groupby("severity", as_index=False).sum()
            result: dict = {}
            result[IMG_NAME] = self._normalize_image_name(base_info[IMG_NAME])
            result["targets"] = targets
            result["critical"] = cve_count_by_status[
                cve_count_by_status.severity == "CRITICAL"
            ].values[0][1]
            result["high"] = cve_count_by_status[
                cve_count_by_status.severity == "HIGH"
            ].values[0][1]
            result["medium"] = cve_count_by_status[
                cve_count_by_status.severity == "MEDIUM"
            ].values[0][1]
            result["low"] = cve_count_by_status[
                cve_count_by_status.severity == "LOW"
            ].values[0][1]
            result["image_link"] = f"../tmp/{result[IMG_NAME]}_report.html"

            image_infos.append(result)
        content = self.general_template.render(imageinfos=image_infos)
        report_path = Path(self.output_path) / f"general_report.html"
        try:
            with open(report_path.as_posix(), "w") as f:
                f.write(content)
        except FileNotFoundError:
            print(f"Can't create report file {report_path}, make sure the {self.output_path} directory exists.")
            sys.exit(1)

    def append_general_report(self, image_info: dict, cve_list: List[dict]):
        for cve in cve_list:
            vul_df: pd.DataFrame = pd.DataFrame(cve[TARGET_VULNERABILITY_DF])
            if len(vul_df) != 0:
                cve_count_by_status = (
                    vul_df.groupby(["severity", "status"])
                    .size()
                    .reset_index(name="CVE_Count")
                )
                vul_count_html = cve_count_by_status.to_html(index=False)
                cve["target_vul_html"] = vul_count_html
        content = self.vul_simple_template.render(
            image_info=image_info,
            cve_list=cve_list,
            image_link=f"../tmp/{image_info[IMG_NAME]}_report.html",
        )
        report_path = Path(self.output_path) / f"general_report.html"
        with open(report_path.as_posix(), "a") as f:
            f.write(content)

    def close_general_report(self):
        content = self.end_general_template.render()
        report_path = Path(self.output_path) / f"general_report.html"
        with open(report_path.as_posix(), "a") as f:
            f.write(content)

    def render_template_to_pdf(self):
        # allows to run without platwright installed if export_pdf = False
        try:
            # Third party imports
            from playwright.sync_api import sync_playwright
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Playwright needs to be installed to export to PDF, run : pip install playwright & playwright install"
            )
        with sync_playwright() as p:
            selectors_to_click = ["button"]
            browser = p.chromium.launch()
            page = browser.new_page()
            html_path = Path(self.output_path) / f"general_report.html"
            page.goto(f"file://{html_path.resolve()}")

            page.wait_for_load_state("networkidle")

            # Simule des clics sur les éléments des sections à dérouler (via leurs sélecteurs CSS)
            for selector in selectors_to_click:
                page.click(selector)

            pdf_path = Path(self.output_path) / f"general_report.pdf"
            page.pdf(path=pdf_path, format="A4", print_background=True)

            browser.close()

    @staticmethod
    def _normalize_image_name(image_name: str) -> str:
        """
        This function normalize the image name repo_name/image_name to repo_name-image_name
        :param image_name:
        :return:
        """
        return image_name.replace("/", "-")



"""
cve_evaluator

===========

This module provides a class and functions to calculate the `cvss_score` based on the `cvss_vector` of a `CVE`.

Author:
    Pengfei liu

Date:
    2025-02-25
"""
from typing import Dict, List


def parse_cvss_vector(cvss_vector: str)-> Dict[str, str]:
    """
    This function parses the cvss vector string into a cvss vector dictionary. For example, a vector string
     "AV:N/AC:L/PR:N/UI:R/S:U/C:H/I:H/A:H" will be converted to a dictionary like {"AV":"N", "AC","L"}.

    :param cvss_vector: The cvss vector string, e.g. AV:N/AC:L/PR:N/UI:R/S:U/C:H/I:H/A:H
    :return: a dictionary with key as vector type, and value as vector value e.g. {"AV":"N", "AC","L"}
    """
    metrics = cvss_vector.split("/")
    metric_dict = {metric.split(":")[0]: metric.split(":")[1] for metric in metrics}
    return metric_dict


def vector_str_to_value(vector_val_score_map:dict, vector_val: str, vector_type: str) -> float:
    """
    This function converts a vector string into a severity score (float value). The input vector_val_score_map must
    be tested for validly. We don't want to test it each time when this function is called. If nothing found return 0
    :param vector_val_score_map:
    :param vector_val: The enum value of the vector, each vector type has its own enum value
    :param vector_type: The type of the CVSS vector, e.g. AV->Attack Vector, A->Availability Impact
    :return: A score value between 0 and 1, the higher numbers representing a higher degree of severity
    """

    return vector_val_score_map[vector_type].get(vector_val, 0.0)

def check_vector_score_mapping_validity(vector_value_score_map:dict)-> (bool,str):
    """
    This function checks whether a vector value score mapping is valid. The input map must contain all required vector
    types, and each vector type must contain all required vector values for the vector type.
    :param vector_value_score_map: The input vector value score mapping, if a user does not provide one, use the default
    :return: A tuple of result and comments
    """
    expected_vector_types = {"AV":['N', 'A', 'L', 'P'],
                             "AC": ['L', 'H'],
                             "PR": ['N', 'L', 'H'],
                             "UI": ['N', 'R'],
                             "S": ['U', 'C'],
                             "C": ['H', 'L', 'N'],
                             "I": ['H', 'L', 'N'],
                             "A": ['H', 'L', 'N']}

    actual_vector_types = vector_value_score_map.keys()
    if vector_value_score_map is None:
        return False, "The provided vector_value_score_map is empty"
    else:
        # 1. check the map contains all the required vector type
        missing_vec_type = [vec_type for vec_type in list(expected_vector_types.keys()) if vec_type not in actual_vector_types]
        if missing_vec_type:
            return False, f"Missing required vector type: {missing_vec_type}"
        else:
            # 2. check for each vector type, it contains the required values
            for vec_type in actual_vector_types:
                expected_vec_vals:List[str] = list(expected_vector_types[vec_type])
                missing_vec_val = [vec_val for vec_val in expected_vec_vals if vec_val not in list(vector_value_score_map[vec_type].keys())]
                if missing_vec_val:
                    return False, f"Vector type {vec_type} is missing required value: {missing_vec_val}"
            return True, "The provided vector_value_score_map is valid"






class CveEvaluator:
    """
    Represents a cve evaluator in the system.
    This class calculates the basic scores with a given CVSS vector. The equation is from
    https://www.first.org/cvss/v3.1/specification-document. You can find all equation details in
    section 7.1

    :param cvss_vector: The cvss vector of the cve
    :type cvss_vector: str


    :ivar cvss_vector: The cvss vector of the cve in string format
    :vartype cvss_vector: str
    :ivar cvss_vec_dict: vss vector of the cve in dictionary format
    :vartype cvss_vec_dict: Dict[str, str]
    :ivar cvss_score: The cvss score of the cve
    :vartype cvss_score: float

    """

    def __init__(self, cvss_vector: str):
        """
        Initializes a cve evaluator
        :param cvss_vector: The cvss vector of the cve
        """
        self.cvss_vector:str = cvss_vector
        self.cvss_vec_dict:Dict[str,str] = parse_cvss_vector(cvss_vector)
        self.cvss_score:Dict[str, float] = self.get_cvss_score()

    def get_cvss_score(self) -> Dict[str, float]:
        """
        This function converts a cvss vector string into a cvss score dictionary.
        :return:
        """
        # get the vector value dict
        vectors = parse_cvss_vector(self.cvss_vector)
        # transform the vector value in str to float
        return {
            vec_type: vector_str_to_value(vec_val, vec_type)
            for vec_type, vec_val in vectors.items()
        }

    def get_exploit_score(self) -> float:
        """
        This function calculates the exploitability score of a CVSS vector.
        :return:
        """
        av = self.cvss_score.get("AV", 0.0)
        ac = self.cvss_score.get("AC", 0.0)
        pr = self.cvss_score.get("PR", 0.0)
        ui = self.cvss_score.get("UI", 0.0)

        return 8.22 * av * ac * pr * ui

    def get_impact_score(self) -> float:
        """
        This function calculates the impact score of a CVSS vector.
        :return:
        """
        c = self.cvss_score.get("C", 0.0)
        i = self.cvss_score.get("I", 0.0)
        a = self.cvss_score.get("A", 0.0)
        # if scope is unchanged
        if self.cvss_score.get("S") == "U":
            return 1 - (1 - c) * (1 - i) * (1 - a)
        else:
            return 7.52 * (1 - (1 - c) * (1 - i) * (1 - a)) - 3.25

    def get_cvss_base_score(self)->float:
        """
        This function calculates the cvss base score of a CVSS vector.
        :return:
        """
        impact = self.get_impact_score()
        exploit = self.get_exploit_score()
        # if scope is unchanged
        if self.cvss_score.get("S") == "U":
            return round(min(10, (impact + exploit)))
        else:
            return round(min(10, 1.08 * (impact + exploit)))



import pandas as pd
import pytest
from importlib import resources
from iv_sherlock.conf.constant import OS_FAMILY, REPO_TAGS, IMG_ID, OS_NAME, OS_ARCHI, IMG_NAME, IMG_CREATION_TIME
from iv_sherlock.image_vul_parser import ImageVulParser


@pytest.fixture
def redis_img_vul_parser():
    report_path = str(resources.files("iv_sherlock.data") / "redis.json")
    return ImageVulParser(raw_report_path=report_path)


def test_getRawData(redis_img_vul_parser):
    raw_data = redis_img_vul_parser.get_raw_data()
    print(raw_data)

def test_parseImageMetadata(redis_img_vul_parser):
    expected_os_family = "debian"
    expected_os_name = "12.7"
    expected_os_archi = "amd64"
    expected_repo_tags = ['redis:latest']
    expected_image_id = "sha256:590b81f2fea1af9798db4580a6299dafba020c2f5dc7d8d734663e7fa5299ca0"
    res = redis_img_vul_parser.parse_image_metadata()
    os_family = res[OS_FAMILY]
    os_name = res[OS_NAME]
    os_archi = res[OS_ARCHI]
    repo_tags = res[REPO_TAGS]
    image_id = res[IMG_ID]
    print(f"os_family: {os_family}")
    print(f"os_name: {os_name}")
    print(f"os_archi: {os_archi}")
    print(f"repo_tags: {repo_tags}")
    print(f"image_id: {image_id}")
    assert os_family == expected_os_family
    assert os_name == expected_os_name
    assert os_archi == expected_os_archi
    assert repo_tags == expected_repo_tags
    assert image_id == expected_image_id

def test_getImageBasicInfo(redis_img_vul_parser):
    expected_os_family = "debian"
    expected_os_name = "12.7"
    expected_os_archi = "amd64"
    expected_repo_tags = ['redis:latest']
    expected_image_id = "sha256:590b81f2fea1af9798db4580a6299dafba020c2f5dc7d8d734663e7fa5299ca0"
    expected_image_name = "redis"
    expected_image_creation_time = "2024-09-13T09:13:14.153929471+02:00"
    res = redis_img_vul_parser.get_image_basic_info()
    os_family = res[OS_FAMILY]
    os_name = res[OS_NAME]
    os_archi = res[OS_ARCHI]
    repo_tags = res[REPO_TAGS]
    image_id = res[IMG_ID]
    image_name = res[IMG_NAME]
    image_creation_time = res[IMG_CREATION_TIME]
    print(res)
    assert os_family == expected_os_family
    assert os_name == expected_os_name
    assert os_archi == expected_os_archi
    assert repo_tags == expected_repo_tags
    assert image_id == expected_image_id
    assert image_creation_time == expected_image_creation_time
    assert image_name == expected_image_name



def test_buildImageVulReport(redis_img_vul_parser):
    img_info, cve_list = redis_img_vul_parser.build_image_vul_report()
    for cve in cve_list:
        print(cve)

def test_parserTargetVulnerabilities(redis_img_vul_parser):
    vuls = redis_img_vul_parser.raw_data.get("Results")[0].get("Vulnerabilities")

    pdf = redis_img_vul_parser.parser_target_vulnerabilities(vuls)
    for column in pdf.columns:
        print(f"Column: {column}, Type: {pdf[column].dtype}")
    pd.set_option('display.max_columns', None)  # Show all columns
    pd.set_option('display.max_colwidth', None)  # Show full content in each column
    print(pdf.head())

def test_parserCvss_withValidCvssValue(redis_img_vul_parser):
    expected_v3_score = 3.7
    expected_v3_vector = "CVSS:3.1/AV:N/AC:H/PR:N/UI:N/S:U/C:N/I:L/A:N"
    cvss = redis_img_vul_parser.raw_data.get("Results")[0].get("Vulnerabilities")[0].get("CVSS")
    v3_score,v3_vector = redis_img_vul_parser.parse_cvss(cvss)
    print(v3_score)
    print(v3_vector)
    assert v3_score == expected_v3_score
    assert v3_vector == expected_v3_vector


def test_parserCvss_withBadCvssValue(redis_img_vul_parser):
    cvss = {}
    v3_score,v3_vector = redis_img_vul_parser.parse_cvss(cvss)
    print(v3_score)
    print(v3_vector)
    assert v3_score is None
    assert v3_vector is None

def test_getAvailableAttributes(redis_img_vul_parser):
    expected_attributes = ['SchemaVersion', 'CreatedAt', 'ArtifactName', 'ArtifactType', 'Metadata', 'Results']
    actual_attributes = redis_img_vul_parser.get_available_attributes()
    print(actual_attributes)
    assert expected_attributes == actual_attributes




import pytest

from iv_sherlock.cve_evaluator import CveEvaluator, parse_cvss_vector, vector_str_to_value, \
    check_vector_score_mapping_validity


@pytest.fixture(scope='module')
def cveEvaluator():
    vector = "AV:N/AC:L/PR:N/UI:R/S:U/C:H/I:L/A:H"
    return CveEvaluator(vector)

@pytest.fixture(scope='module')
def cveEvaluatorWithBadVector():
    vector = "AV:N/AC:L/PR:N/UI:R/S:U/C:H/I:L/AO:H"
    return CveEvaluator(vector)


def test_parseCvssVector():
    expected_vector = {'AV': 'N', 'AC': 'L', 'PR': 'N', 'UI': 'R', 'S': 'U', 'C': 'H', 'I': 'H', 'A': 'H'}
    vector = "AV:N/AC:L/PR:N/UI:R/S:U/C:H/I:H/A:H"
    resu = parse_cvss_vector(vector)
    print(resu)
    assert resu == expected_vector


def test_vectorStrToValue():
    expected_score = {'AV': 0.85, 'AC': 0.77, 'PR': 0.85, 'UI': 0.62, 'S': 1.0, 'C': 0.56, 'I': 0.56, 'A': 0.56}
    cvss_score = {}
    vector = {'AV': 'N', 'AC': 'L', 'PR': 'N', 'UI': 'R', 'S': 'U', 'C': 'H', 'I': 'H', 'A': 'H'}
    for vec_type, vec_val in vector.items():
        print(f"type: {vec_type}, val: {vec_val}")
        vec_score = vector_str_to_value(vec_val,vec_type)
        print(f"type: {vec_type}, val: {vec_score}")
        cvss_score[vec_type] = vec_score
    print(cvss_score)
    assert cvss_score == expected_score

def test_CveEvaluator_getCvssScore(cveEvaluator):
    expected_score = {'AV': 0.85, 'AC': 0.77, 'PR': 0.85, 'UI': 0.62, 'S': 1.0, 'C': 0.56, 'I': 0.22, 'A': 0.56}
    cvss_score = cveEvaluator.get_cvss_score()
    print(cvss_score)
    assert cvss_score == expected_score

def test_CveEvaluator_getExploitScore(cveEvaluator):
    expected_score = 2.835
    exploit_score = cveEvaluator.get_exploit_score()
    resu = round(exploit_score, 3)
    print(exploit_score)
    assert resu == expected_score

def test_CveEvaluator_getImpactScore(cveEvaluator):
    expected_score = 3.134
    impact_score = cveEvaluator.get_impact_score()
    print(impact_score)
    resu = round(impact_score, 3)
    assert resu == expected_score

def test_CveEvaluator_getCvssBaseScore_withValidVector(cveEvaluator):
    expected_base_score = 6
    actual_base_score = cveEvaluator.get_cvss_base_score()
    print(f"actual base score: {actual_base_score}")
    assert expected_base_score == actual_base_score


def test_CveEvaluator_getCvssBaseScore_withBadVectorType(cveEvaluatorWithBadVector):
    expected_base_score = 6
    actual_base_score = cveEvaluatorWithBadVector.get_cvss_base_score()
    print(f"actual base score: {actual_base_score}")
    assert expected_base_score == actual_base_score

def test_checkVectorScoreMappingValidity_withValidVector(cveEvaluator):
    vector_score_mapping = {
        "AV": {"N": 0.85, "A": 0.62, "L": 0.55, "P": 0.2},
        "AC": {"L": 0.77, "H": 0.44},
        "PR": {"N": 0.85, "L": 0.62, "H": 0.27},
        "UI": {"N": 0.85, "R": 0.62},
        "S": {"U": 1.0, "C": 1.08},
        "C": {"H": 0.56, "L": 0.22, "N": 0.0},
        "I": {"H": 0.56, "L": 0.22, "N": 0.0},
        "A": {"H": 0.56, "L": 0.22, "N": 0.0},
    }
    resu,comments = check_vector_score_mapping_validity(vector_score_mapping)
    print(comments)
    assert resu == True

def test_checkVectorScoreMappingValidity_withMissVectorType(cveEvaluator):
    vector_score_mapping = {
        "AV": {"N": 0.85, "A": 0.62, "L": 0.55, "P": 0.2},
        "AC": {"L": 0.77, "H": 0.44},
        "PR": {"N": 0.85, "L": 0.62, "H": 0.27},
        "C": {"H": 0.56, "L": 0.22, "N": 0.0},
        "I": {"H": 0.56, "L": 0.22, "N": 0.0},
    }
    resu, comments = check_vector_score_mapping_validity(vector_score_mapping)
    print(comments)
    assert resu == False

def test_checkVectorScoreMappingValidity_withMoreVectorType(cveEvaluator):
    vector_score_mapping = {
        "AV": {"N": 0.85, "A": 0.62, "L": 0.55, "P": 0.2},
        "AC": {"L": 0.77, "H": 0.44},
        "PR": {"N": 0.85, "L": 0.62, "H": 0.27},
        "UI": {"N": 0.85, "R": 0.62},
        "S": {"U": 1.0, "C": 1.08},
        "C": {"H": 0.56, "L": 0.22, "N": 0.0},
        "I": {"H": 0.56, "L": 0.22, "N": 0.0},
        "A": {"H": 0.56, "L": 0.22, "N": 0.0},
        "ATEST": {"H": 0.56, "L": 0.22, "N": 0.0},
    }
    resu, comments = check_vector_score_mapping_validity(vector_score_mapping)
    print(comments)
    assert resu == True

def test_checkVectorScoreMappingValidity_withMissVectorVal(cveEvaluator):
    vector_score_mapping = {
        "AV": {"N": 0.85, "A": 0.62,},
        "AC": {"L": 0.77, "H": 0.44},
        "PR": {"N": 0.85, "L": 0.62, "H": 0.27},
        "UI": {"N": 0.85, "R": 0.62},
        "S": {"U": 1.0, "C": 1.08},
        "C": {"H": 0.56, "L": 0.22, "N": 0.0},
        "I": {"H": 0.56, "L": 0.22, "N": 0.0},
        "A": {"H": 0.56, "L": 0.22, "N": 0.0},
    }
    resu,comments = check_vector_score_mapping_validity(vector_score_mapping)
    print(comments)
    assert resu == False


# Standard library imports
import argparse
import glob
from pathlib import Path

# sherlock imports

from iv_sherlock.config_manager import Config
from iv_sherlock.image_vul_parser import ImageVulParser
from iv_sherlock.report_generator import ReportGenerator


def main():


    # create an arg parse for cli
    parser = argparse.ArgumentParser(description="Parse raw vulnerabilities reports and generate human readable report.")
    parser.add_argument("-c","--conf",  type=str, help="The custom configuration file path. The configuration file must be in .toml format.")
    args = parser.parse_args()
    # create a config manager
    config = Config(args.conf)
    # get values from config manager
    raw_data_root_dir = config.get_input_report_path()

    export_pdf = config.get_export_pdf()
    # store results
    results = []
    for image_path in glob.glob(f"{raw_data_root_dir}/*.json"):
        raw_data_path = Path(image_path)
        # create image vul parser
        img_vul_parser = ImageVulParser(raw_data_path.as_posix())
        results.append(img_vul_parser.build_image_vul_report())
    if len(results) == 0:
        print(f"Can't find any vulnerabilities report. Make sure the report source directory {config.get_input_report_path()} is correct.")
    # create report generator
    rg = ReportGenerator(config)
    rg.initiate_general_report(results)
    for base_info, cve_list in results:
        rg.generate_report(base_info, cve_list)
        rg.append_general_report(image_info=base_info, cve_list=cve_list)

    rg.close_general_report()
    if export_pdf:
        rg.render_template_to_pdf()


if __name__ == "__main__":
    main()



import json
import tomllib  # Use 'import toml' if using Python < 3.11
from importlib import resources


class Config:
    _instance = None  # Singleton instance

    def __new__(cls, custom_config_path: str = None):
        """
        Initiate a Config instance and ensure only one instance exists (Singleton).
        :param custom_config_path:
        """
        if cls._instance is None:
            cls._instance = super(Config, cls).__new__(cls)
            # load the default config
            cls._instance._default_config = tomllib.loads(resources.read_text("iv_sherlock.conf", "config.toml"))
            # create custom config
            cls._instance._custom_config={}
        # if a custom config file is provided, load the custom config
        if custom_config_path:
            cls._instance._custom_config = cls.load_custom_config(custom_config_path)
        # merge the default and custom config
        cls._instance._config = cls.merge_dicts(cls._instance._default_config, cls._instance._custom_config)
        return cls._instance

    @staticmethod
    def load_custom_config(config_path:str):
        """
        Loads a TOML config file and return the corresponding config dictionary.
        :param config_path: Input config file path.
        :return:
        """
        custom_config = {}
        try:
            with open(config_path, "rb") as f:
                custom_config = tomllib.load(f)
        except FileNotFoundError:
            print(f"Warning: User config file {config_path} not found. Ignoring user config.")
        except Exception as e:
            print(f"Error loading config file {config_path}: {e}")
        return custom_config

    @staticmethod
    def merge_dicts(default_config: dict, custom_config: dict) -> dict:
        """
        Recursively merge two dictionaries, keeping default values if custom values are missing or empty.
        :param default_config: A dict contains the default values
        :param custom_config: A dict contains the custom values
        :return:
        """
        merged_config = default_config.copy()  # Start with default values

        for key, value in custom_config.items():
            # Recursively merge nested dicts
            if isinstance(value, dict) and key in merged_config:
                merged_config[key] = Config.merge_dicts(merged_config[key], value)
            # Ignore empty string values, keeping the default
            elif value is None or value == "" or value == " ":
                continue
            # Override with non-empty values
            else:
                merged_config[key] = value
        return merged_config

    def get_input_report_path(self)->str:
        return self._config["report"]["source_path"]

    def get_output_report_path(self)->str:
        return self._config["report"]["out_path"]

    def get_export_pdf(self)->bool:
        return self._config["report"]["export_pdf"]

    def get_default_encoding(self)->str:
        return self._config["report"]["default_encoding"]

    def get_cvss_vector_value_mapping(self)->dict:
        score_map_str = self._config["cvss"]["score_mapping"]
        return json.loads(score_map_str)

    @staticmethod
    def get_report_template_dir_path()->str:
        """
        This method returns the path to the output report template directory.
        :return: Path to the output report template directory
        """
        return str(resources.files("iv_sherlock") / "templates")






# Initialize the config singleton on first import
config = Config()
